// recognizer.h

#ifdef _MSC_VER
#define NOMINMAX
#endif

#include "recognizer1.h"

namespace kaldi {
namespace nnet1 {
NnetRecognizer::NnetRecognizer(
        const kaldi::nnet1::Nnet& nnet,
        kaldi::nnet1::PdfPrior& pdf_prior,
        const kaldi::TransitionModel& trans_model,
        kaldi::nnet1::DecodableNnetOnlineForHarkOptions opt,
        std::shared_ptr<kaldi::OnlineFeatureMfcnet>& features,
        std::shared_ptr<kaldi::OnlineFeatureInterface>& features_if,
        const fst::Fst<fst::StdArc>* decode_fst,
        const kaldi::LatticeFasterDecoderConfig& decoder_config,
        const fst::SymbolTable* phone_syms,
        const fst::SymbolTable* word_syms,
        const kaldi::WordAlignLatticeLexiconInfo& lexicon_info,
        std::string& lm_name,
        int emulate_julius,
        int nbest_count,
        int progress_output_interval,
        bool enable_print_result,
        bool enable_print_progress,
        bool enable_print_sentence,
        std::shared_ptr<result_server> result_client):
    id_(features->GetID()),
    trans_model_(trans_model),
    features_(features), features_if_(features_if),
    decodable_(nnet, pdf_prior, trans_model, opt, features_if),
    decoder_(*decode_fst, decoder_config),
    phone_syms_(phone_syms), word_syms_(word_syms),
    lexicon_info_(lexicon_info),
    result_client_(result_client),
    lm_name_(lm_name),
    emulate_julius_(emulate_julius),
    nbest_count_(nbest_count),
    progress_output_interval_(progress_output_interval),
    enable_print_result_(enable_print_result),
    enable_print_progress_(enable_print_progress),
    enable_print_sentence_(enable_print_sentence),
    features_input_is_finished(false){
  //if(features_){ std::string id_ = features_->GetID(); }
  // remove path prefix
#ifdef _MSC_VER
  std::string::size_type pos = lm_name.rfind("\\");
#else
  std::string::size_type pos = lm_name.rfind("/");
#endif
  lm_name_ = (pos != std::string::npos) ? lm_name.substr(++pos) : lm_name;
  emulate_julius_ = (emulate_julius != 0) ? 1 : 0;
  delimiter_sym_ = (emulate_julius_) ? ".\n" : "";
  //DEBUG//std::cout << "fst ::: " << (decode_fst ? "not NULL" : "NULL") << std::endl;
}

std::shared_ptr<OnlineFeatureInterface> NnetRecognizer::GetOnlineFeatureInterface()
{
  return features_if_;
}

void NnetRecognizer::Recognize()
{
  
  //DEBUG//std::cout << "frames ::: " <<  features_if_->NumFramesReady() << std::endl;
  //DEBUG//std::cout << "decoder ::: " <<  decoder_.NumFramesDecoded() << std::endl;
  //DEBUG//std::cout << "decodable ::: " << decodable_.NumFramesReady() << std::endl;
  if(features_){
    features_->Initialize();
    std::ostringstream ss_sock;
    if(features_->is_ex_mfcnet_input()){
      ss_sock << std::fixed << std::setprecision(6)
              << "<SOURCEINFO CLIENTID=\"" << features_->GetUtteranceInfo().client_id
              <<          "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
              <<          "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
              <<             "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
              <<           "\" AZIMUTH=\"" << features_->GetUtteranceInfo().azimuth
              <<         "\" ELEVATION=\"" << features_->GetUtteranceInfo().elevation
              <<               "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
              <<              "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
              <<               "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
              <<              "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
              <<              "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_
              << "<STARTRECOG SOURCEID=\"" << features_->GetUtteranceInfo().source_id
              <<               "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
              <<              "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
              <<              "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_;
    }
    //else if(features_->is_file_input() || features_->is_mfcnet_input()){
    else{
      ss_sock << std::fixed << std::setprecision(6)
              << "<SOURCEINFO SOURCEID=\"" << features_->GetSourceInfo().source_id
              <<           "\" AZIMUTH=\"" << features_->GetSourceInfo().azimuth
              <<         "\" ELEVATION=\"" << features_->GetSourceInfo().elevation
              <<               "\" SEC=\"" << features_->GetSourceInfo().tv_sec
              <<              "\" USEC=\"" << features_->GetSourceInfo().tv_usec << "\" />\n" << delimiter_sym_
              << "<STARTRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    }
    result_client_->broadcast(ss_sock.str());
  }

  //DEBUG//KALDI_LOG << "started recognize key id=" << id_;
  Timer timer; timer.Reset();
  //decoder_.Decode(&decodable_);
  decoder_.InitDecoding();

  int progress_iframes = 0;
  //int progress_dframes = progress_output_interval_;
  int progress_dframes = 0;
  int progress_wait = static_cast<int>(progress_output_interval_ * 0.5f);
  while(true){
    if(features_){
      //DEBUG//KALDI_LOG << "receiving features key id=" << id_;
      features_->Receive();
    }
    if(features_if_->IsLastFrame(features_if_->NumFramesReady() - 1)){
      decoder_.AdvanceDecoding(&decodable_);
      //DEBUG//KALDI_LOG << "final frame recognized key id=" << id_;
      break;
    }
    if(progress_output_interval_ > 0){
      bool not_enough = true;
      while(features_if_->NumFramesReady() > progress_iframes){
        not_enough = false;
        decoder_.AdvanceDecoding(&decodable_, progress_output_interval_);
        progress_iframes += progress_output_interval_;
        if(decoder_.NumFramesDecoded() > progress_dframes){
          progress_dframes = decoder_.NumFramesDecoded();
          WriteIntermediateResult(progress_dframes);
        }
      }
      if(not_enough){
        multiplatform::time::ms_sleep(progress_wait);
      }
    }
    else{
      multiplatform::time::ms_sleep(500);
    }
  }

  decoder_.FinalizeDecoding();
  //DEBUG//KALDI_LOG << "finalized decoding key id=" << id_;
  double elapsed = timer.Elapsed();
  KALDI_LOG << elapsed << " sec, RT=" << (elapsed * 100.0 / features_if_->NumFramesReady());

  if(features_){
    std::ostringstream ss_sock;
    if(features_->is_ex_mfcnet_input()){
      ss_sock << "<ENDRECOG CLIENTID=\"" << features_->GetUtteranceInfo().client_id
              <<        "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
              <<        "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
              <<           "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
              <<             "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
              <<            "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
              <<            "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_;
    }
    //else if(features_->is_file_input() || features_->is_mfcnet_input()){
    else{
      ss_sock << "<ENDRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    }
    result_client_->broadcast(ss_sock.str());
  }
  WriteResult();

  if(features_){
    result_client_->add_recogend_key(id_);
  }
}

void NnetRecognizer::WriteIntermediateResult(kaldi::int32 frame)
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0){
    std::ostringstream ss_sock;
    if(features_->is_ex_mfcnet_input()){
      ss_sock << "<RECOGFAIL CLIENTID=\"" << features_->GetUtteranceInfo().client_id
              <<         "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
              <<         "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
              <<            "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
              <<              "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
              <<             "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
              <<             "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_;
    }
    //else if(features_->is_file_input() || features_->is_mfcnet_input()){
    else{
      ss_sock << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
    }
    result_client_->broadcast(ss_sock.str());
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat, false);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned &weight);

  {
    std::ostringstream ss_sock;
    if(features_->is_ex_mfcnet_input()){
      ss_sock << "<RECOGOUT CLIENTID=\"" << features_->GetUtteranceInfo().client_id
              <<        "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
              <<        "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
              <<           "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
              <<             "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
              <<            "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
              <<            "\" UUID=\"" << features_->GetID() << "\"";
    }
    //else if(features_->is_file_input() || features_->is_mfcnet_input()){
    else{
      ss_sock << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
    }
    if(lm_name_ != ""){ ss_sock << " NAME=\"" << lm_name_ << "\""; }
    ss_sock << std::fixed << std::setprecision(6)
            << ">\n"
            << " <PHYPO RANK=\"" << 1
            <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
            <<   "\" AMSCORE=\"" << -(double)weight.Value2()
            <<   "\" LMSCORE=\"" << -(double)weight.Value1()
            <<     "\" FRAME=\"" << frame
            <<      "\" TIME=\"" << time(NULL)
            << "\">\n";
    size_t i = 0;
    //int word_index = 0;
    while(i < words.size()){
      std::string classid = word_syms_->Find(words[i]);
      std::string word = string_tokenizer(classid, "+")[0];
      //std::string word = (word_index < string_tokenizer(classid, "+").size()) ? string_tokenizer(classid, "+")[word_index] : string_tokenizer(classid, "+")[0];
      if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss_sock << std::fixed << std::setprecision(3)
              << "  <WHYPO WORD=\"" << word
              <<    "\" CLASSID=\"" << classid
              <<      "\" PHONE=\"" << ""
              <<         "\" CM=\"" << 1.0f << "\"/>\n";
      i++;
    }
    ss_sock << " </PHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

    result_client_->broadcast(ss_sock.str());
  }
}

void NnetRecognizer::WriteResult(std::ostream& out)
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0){
    // Write error result for socket
    if(features_){
      std::ostringstream ss_sock;
      if(features_->is_ex_mfcnet_input()){
        ss_sock << "<RECOGFAIL CLIENTID=\"" << features_->GetUtteranceInfo().client_id
                <<         "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
                <<         "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
                <<            "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
                <<              "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
                <<             "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
                <<             "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_;
      }
      //else if(features_->is_file_input() || features_->is_mfcnet_input()){
      else{
        ss_sock << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
      }
      result_client_->broadcast(ss_sock.str());
    }

    // Write error result for stdout
    if(enable_print_result_ && enable_print_sentence_){
      std::stringstream ss_cout;
      if(emulate_julius_){
        if(features_){
          if(features_->is_ex_mfcnet_input()){
            ss_cout << std::fixed << std::setprecision(6)
                    <<   "client_id = " << features_->GetUtteranceInfo().client_id
                    << ", mic_index = " << features_->GetUtteranceInfo().mic_index
                    << ", source_id = " << features_->GetUtteranceInfo().source_id
                    <<    ", sub_id = " << features_->GetUtteranceInfo().sub_id
                    <<   ", azimuth = " << features_->GetUtteranceInfo().azimuth
                    << ", elevation = " << features_->GetUtteranceInfo().elevation
                    <<       ", sec = " << features_->GetUtteranceInfo().tv_sec
                    <<      ", usec = " << features_->GetUtteranceInfo().tv_usec
                    <<      ", uuid = " << features_->GetID() << "\n";
          }
          //else if(features_->is_file_input() || features_->is_mfcnet_input()){
          else{
            ss_cout << std::fixed << std::setprecision(6)
                    <<   "source_id = " << features_->GetSourceInfo().source_id
                    <<   ", azimuth = " << features_->GetSourceInfo().azimuth
                    << ", elevation = " << features_->GetSourceInfo().elevation
                    <<       ", sec = " << features_->GetSourceInfo().tv_sec
                    <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
          }
        }
        else{
          ss_cout << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n";
        }
        ss_cout << "### Recognition: 2nd pass (RL heuristic best-first) \n"
                << "STAT: 00 \n"
                << "sentence1: \n"
                << "wseq1: \n"
                << "phseq1: \n"
                << "cmscore1: \n"
                << "score1: \n\n"
                << "ERROR: recognition failed.";
      }
      else{
        ss_cout << "recognition failed\n"
                << "result: \n"
                << "phone_seq: ";
      }
      out << ss_cout.str() << std::endl;
    }

    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned &weight);

  {
    std::vector<std::string> str_classids; str_classids.resize(words.size());
    std::vector<std::string> str_words; str_words.resize(words.size());

    if(word_syms_ != NULL){
      size_t i = 0;
      //int word_index = 0;
      while(i < words.size()){
        str_classids[i] = word_syms_->Find(words[i]);
        str_words[i] = string_tokenizer(str_classids[i], "+")[0];
        //str_words[i] = (word_index < string_tokenizer(str_classids[i], "+").size()) ? string_tokenizer(str_classids[i], "+")[word_index] : string_tokenizer(str_classids[i], "+")[0];
        if(str_classids[i] == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
        i++;
      }
    }
    else{
      size_t i = 0;
      while(i < words.size()){
        str_classids[i] = str_words[i] = std::to_string(words[i]);
        i++;
      }
    }

    // Write result for socket
    if(features_){
      std::ostringstream ss_sock;

      if(features_->is_ex_mfcnet_input()){
        ss_sock << "<RECOGOUT CLIENTID=\"" << features_->GetUtteranceInfo().client_id
                <<        "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
                <<        "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
                <<           "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
                <<             "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
                <<            "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
                <<            "\" UUID=\"" << features_->GetID() << "\"";
      }
      //else if(features_->is_file_input() || features_->is_mfcnet_input()){
      else{
        ss_sock << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
      }
      if(lm_name_ != ""){ ss_sock << " NAME=\"" << lm_name_ << "\""; }
      ss_sock << std::fixed << std::setprecision(6)
              << ">\n"
              << " <SHYPO RANK=\"" << 1
              <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
              <<   "\" AMSCORE=\"" << -(double)weight.Value2()
              <<   "\" LMSCORE=\"" << -(double)weight.Value1() << "\">\n";
      size_t i = 0;
      while(i < words.size()){
        ss_sock << std::fixed << std::setprecision(3)
                << "  <WHYPO WORD=\"" << str_words[i]
                <<    "\" CLASSID=\"" << str_classids[i]
                <<      "\" PHONE=\"" << ""
                <<         "\" CM=\"" << 1.0f << "\"/>\n";
        i++;
      }
      ss_sock << " </SHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

      result_client_->broadcast(ss_sock.str());
    }

    // Write result for stdout
    {
      std::stringstream ss_cout;

      if(enable_print_result_ && enable_print_sentence_){
        if(emulate_julius_){
          if(features_){
            if(features_->is_ex_mfcnet_input()){
              ss_cout << std::fixed << std::setprecision(6)
                      <<   "client_id = " << features_->GetUtteranceInfo().client_id
                      << ", mic_index = " << features_->GetUtteranceInfo().mic_index
                      << ", source_id = " << features_->GetUtteranceInfo().source_id
                      <<    ", sub_id = " << features_->GetUtteranceInfo().sub_id
                      <<   ", azimuth = " << features_->GetUtteranceInfo().azimuth
                      << ", elevation = " << features_->GetUtteranceInfo().elevation
                      <<       ", sec = " << features_->GetUtteranceInfo().tv_sec
                      <<      ", usec = " << features_->GetUtteranceInfo().tv_usec
                      <<      ", uuid = " << features_->GetID() << "\n";
            }
            //else if(features_->is_file_input() || features_->is_mfcnet_input()){
            else{
              ss_cout << std::fixed << std::setprecision(6)
                      <<   "source_id = " << features_->GetSourceInfo().source_id
                      <<   ", azimuth = " << features_->GetSourceInfo().azimuth
                      << ", elevation = " << features_->GetSourceInfo().elevation
                      <<       ", sec = " << features_->GetSourceInfo().tv_sec
                      <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
            }
          }
          else{
            ss_cout << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n";
          }

          ss_cout << "### Recognition: 2nd pass (RL heuristic best-first) \n"
                  << "STAT: 00 \n"
                  << "sentence1: ";
          {
            size_t i = 0;
            while(i < words.size()){ ss_cout << str_words[i++] << " "; }
          }
          ss_cout << "\n";

          ss_cout << "wseq1: ";
          {
            size_t i = 0;
            while(i < words.size()){ ss_cout << str_classids[i++] << " "; }
          }
          ss_cout << "\n";

          ss_cout << "phseq1: ";
          if(phone_syms_ != NULL){
            size_t i = 0;
            while(i < phones_aligned.size()){
              kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
              kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
              std::string s = phone_syms_->Find(phone);
              if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
              ss_cout << s << " ";
              i++;
            }
          }
          ss_cout << "\n";

          ss_cout << "cmscore1: ";
          {
            size_t i = 0;
            while(i < words.size()){
              ss_cout << std::fixed << std::setprecision(3) << 1.0f << " ";
              i++;
            }
          }
          ss_cout << "\n";

          ss_cout << std::fixed << std::setprecision(6)
                  << "score1: " << -(double)(weight.Value1() + weight.Value2())
                  << " ( AM: " << -(double)weight.Value2() << ", LM: " << -(double)weight.Value1() << " )";
        }
        else{
          ss_cout << "result: ";
          {
            size_t i = 0;
            while(i < words.size()){ ss_cout << str_classids[i++] << " "; }
          }
          ss_cout << "\n";

          ss_cout << "phone_seq: ";
          if(phone_syms_ != NULL){
            size_t i = 0;
            while(i < phones_aligned.size()){
              kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
              kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
              std::string s = phone_syms_->Find(phone);
              if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
              ss_cout << s << " ";
              i++;
            }
          }
          ss_cout << "\n";

          ss_cout << std::fixed << std::setprecision(6)
                  <<     "score=" << -(double)(weight.Value1() + weight.Value2())
                  << ", amscore=" << -(double)weight.Value2()
                  << ", lmscore=" << -(double)weight.Value1();
        }

        out << ss_cout.str() << std::endl;
      }
    }
  }
}

} // namespace nnet1
} // namespace kaldi

