// pidstat.cpp: pidstat like test command by C++11 coding, Report statistics for Linux tasks

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <dirent.h>
#include <ctype.h>
#include <sys/types.h>
#include <pwd.h>
#include <sys/utsname.h>
#include <regex.h>
#include <inttypes.h>

#include <sstream>
#include <thread>
#include <atomic>
#include <fstream>

#include "libstats.hpp"
#include "read_stats.hpp"

namespace multiplatform {

//int64_t interval = -1;

int dplaces_nr = -1;    // Number of decimal places

void resource_tracker::usage(char *progname)
{
  std::stringstream ss;
  ss << "Usage:\n  "
     << progname << " [ options ] [ <interval> [ <count> ] ] [ -e <program> <args> ]\n"
        "Options are:\n"
        "  [ -d ] [ -H ] [ -h ] [ -I ] [ -l ] [ -R ] [ -r ] [ -s ] [ -t ]\n"
        "  [ -U [ <username> ] ] [ -u ] [ -V ] [ -v ] [ -w ] [ -C <command> ]\n"
        "  [ -G <process_name> ] [ -p { <pid> [,...] | SELF | ALL } ]\n"
        "  [ -T { TASK | CHILD | ALL } ] [ --dec={ 0 | 1 | 2 } ] [ --human ]\n";
  fprintf(stderr, "%s", ss.str().c_str());
  exit(1);
}

//void resource_tracker::alarm_handler(int sig)
//{
//  alarm(interval);
//}


//void resource_tracker::int_handler(int sig)
//{
//  signal_caught = 1;
//}


void resource_tracker::sfree_pid(struct st_pid **plist, int force)
{
  int i;
  struct st_pid *p;

  while(*plist != NULL){
    p = *plist;
    if(!p->exist || force){
      *plist = p->next;
      for(i = 0; i < 3; i++){
        if(p->pstats[i]){ free(p->pstats[i]); }
      }
      free(p);
    }
    else{ plist = &(p->next); }
  }
}


void resource_tracker::set_pid_nonexistent(struct st_pid *plist)
{
  while(plist != NULL){
    plist->exist = FALSE;
    plist = plist->next;
  }
}


struct resource_tracker::st_pid* resource_tracker::add_list_pid(struct st_pid **plist, pid_t pid, pid_t tgid)
{
  struct st_pid *p, *ps, *tgid_p = NULL;
  int i;
  int tgid_found = FALSE;

  if(!pid){ return NULL; }

  if(!tgid){
    // Add a true PID to the list.
    // Add it in ascending order, not taking into account other TIDs.
    while(*plist != NULL){
      p = *plist;
      // PID found in list
      if(!p->tgid && (p->pid == pid)){ return p; }

      // Stop now to insert PID in list
      if(!p->tgid && (p->pid > pid)){ break; }

      plist = &(p->next);
    }
  }
  else{
    // PID is a TID.
    // It will be inserted in ascending order immediately following its TGID.
    while(*plist != NULL){
      p = *plist;
      if(p->pid == tgid){
        // TGID found in list
        tgid_found = TRUE;
        tgid_p = p;
        break;
      }

      plist = &(p->next);
    }
    // TGID not found: Stop now
    if(!tgid_found){ return NULL; }

    plist = &(p->next);
    while(*plist != NULL){
      p = *plist;
      // TID found in list
      if((p->tgid == tgid_p) && (p->pid == pid)){ return p; }

      // Stop now to insert TID in list
      if((p->tgid == tgid_p) && (p->pid > pid)){ break; }

      // End of TID list: insert TID here
      if(p->tgid != tgid_p){ break; }

      plist = &(p->next);
    }
  }

  // PID not found
  ps = *plist;

  // Add PID to the list
  if((*plist = (struct st_pid *) malloc(sizeof(struct st_pid))) == NULL){
    perror("malloc");
    //exit(4);
    return ps;
  }
  memset(*plist, 0, sizeof(struct st_pid));

  p = *plist;
  for(i = 0; i < 3; i++){
    if((p->pstats[i] = (struct pid_stats *) malloc(sizeof(struct pid_stats))) == NULL){
      perror("malloc");
      //exit(4);
      for(--i; i >= 0; --i){
        free(p->pstats[i]);
      }
      return ps;
    }
    memset(p->pstats[i], 0, sizeof(struct pid_stats));
  }
  p->pid = pid;
  p->next = ps;
  if(tgid_p){ p->tgid = tgid_p; }

  return p;
}


void resource_tracker::read_proc_meminfo(void)
{
  struct stats_memory st_mem;

  memset(&st_mem, 0, sizeof(struct stats_memory));
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->read_meminfo(&st_mem);
    delete res;
  }
  tlmkb = st_mem.tlmkb;
}


int resource_tracker::read_proc_pid_stat(pid_t pid, struct st_pid *plist,
           unsigned int *thread_nr, pid_t tgid, int curr)
{
  constexpr static char PID_STAT[] = "/proc/%u/stat";
  constexpr static char TASK_STAT[] = "/proc/%u/task/%u/stat";
  int fd, sz, rc;
  size_t commsz;
  char filename[128];
  static char buffer[1024 + 1];
  char *start, *end;
  struct pid_stats *pst = plist->pstats[curr];

  if(tgid){ sprintf(filename, TASK_STAT, tgid, pid); }
  else{ sprintf(filename, PID_STAT, pid); }

  // No such process
  if((fd = open(filename, O_RDONLY)) < 0){ return 1; }

  sz = read(fd, buffer, 1024);
  close(fd);
  if(sz <= 0){ return 1; }
  buffer[sz] = '\0';

  if((start = strchr(buffer, '(')) == NULL){ return 1; }
  start += 1;
  if((end = strrchr(start, ')')) == NULL){ return 1; }
  commsz = end - start;
  if(commsz >= MAX_COMM_LEN){ return 1; }
  memcpy(plist->comm, start, commsz);
  plist->comm[commsz] = '\0';
  start = end + 2;

  rc = sscanf(start,
        "%*s %*d %*d %*d %*d %*d %*u"
        " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64
        " %" PRIu64 " %" PRIu64 " %" PRId64 " %" PRId64
        " %*d %*d %u %*u %*d %" PRIu64 " %" PRIu64
        " %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u"
        " %u %u %u %" PRIu64 " %" PRIu64 " %" PRId64 "\n",
        &pst->minflt, &pst->cminflt, &pst->majflt, &pst->cmajflt,
        &pst->utime,  &pst->stime, &pst->cutime, &pst->cstime,
        thread_nr, &pst->vsz, &pst->rss,
        &pst->processor, &pst->priority, &pst->policy,
        &pst->blkio_swapin_delays, &pst->gtime, &pst->cgtime);

  if(rc < 15){ return 1; }

  // gtime and cgtime fields are unavailable in file
  if(rc < 17){ pst->gtime = pst->cgtime = 0; }

  // Convert to kB
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->get_kb_shift();
    pst->vsz >>= 10;
    pst->rss = res->PG_TO_KB(pst->rss);
    delete res;
  }

  return 0;
}


int resource_tracker::read_proc_pid_sched(pid_t pid, struct st_pid *plist, pid_t tgid, int curr)
{
  constexpr static char PID_SCHED[] = "/proc/%u/schedstat";
  constexpr static char TASK_SCHED[] = "/proc/%u/task/%u/schedstat";
  int fd, sz, rc = 0;
  char filename[128];
  static char buffer[1024 + 1];
  uint64_t wtime = 0;
  struct pid_stats *pst = plist->pstats[curr];

  if(tgid){ sprintf(filename, TASK_SCHED, tgid, pid); }
  else{ sprintf(filename, PID_SCHED, pid); }

  if((fd = open(filename, O_RDONLY)) >= 0){
    // schedstat file found for process
    sz = read(fd, buffer, 1024);
    close(fd);
    if(sz > 0){ buffer[sz] = '\0'; rc = sscanf(buffer, "%*u %" PRIu64 " %*d\n", &wtime); }
  }

  // Convert ns to jiffies
  pst->wtime = wtime * hz / 1000000000;

  if(rc < 1){ return 1; }

  return 0;
}


int resource_tracker::read_proc_pid_status(pid_t pid, struct st_pid *plist, pid_t tgid, int curr)
{
  constexpr static char PID_STATUS[] = "/proc/%u/status";
  constexpr static char TASK_STATUS[] = "/proc/%u/task/%u/status";
  FILE *fp;
  char filename[128], line[256];
  struct pid_stats *pst = plist->pstats[curr];

  if(tgid){ sprintf(filename, TASK_STATUS, tgid, pid); }
  else{ sprintf(filename, PID_STATUS, pid); }

  // No such process
  if((fp = fopen(filename, "r")) == NULL){ return 1; }

  while(fgets(line, sizeof(line), fp) != NULL){
    if(!strncmp(line, "Uid:", 4)){
      sscanf(line + 5, "%u", &plist->uid);
    }
    else if(!strncmp(line, "Threads:", 8)){
      sscanf(line + 9, "%u", &pst->threads);
    }
    else if(!strncmp(line, "voluntary_ctxt_switches:", 24)){
      sscanf(line + 25, "%" PRIu64, &pst->nvcsw);
    }
    else if(!strncmp(line, "nonvoluntary_ctxt_switches:", 27)){
      sscanf(line + 28, "%" PRIu64, &pst->nivcsw);
    }
  }

  fclose(fp);

  return 0;
}



int resource_tracker::read_pid_stats(pid_t pid, struct st_pid *plist, unsigned int *thread_nr,
       pid_t tgid, int curr)
{
  if(read_proc_pid_stat(pid, plist, thread_nr, tgid, curr)){ return 1; }

  // No need to test the return code here: Not finding
  // the schedstat files shouldn't make pidstat stop.
  read_proc_pid_sched(pid, plist, tgid, curr);

  if(read_proc_pid_status(pid, plist, tgid, curr)){ return 1; }

  return 0;
}


void resource_tracker::read_task_stats(pid_t pid, struct st_pid *plist, int curr)
{
  constexpr static char PROC_TASK[] = "/proc/%u/task";
  DIR *dir;
  pid_t tid;
  struct dirent *drp;
  char filename[128];
  unsigned int thr_nr;
  struct st_pid *tlist;

  // Open /proc/#/task directory
  sprintf(filename, PROC_TASK, pid);
  if((dir = opendir(filename)) == NULL){ return; }

  while((drp = readdir(dir)) != NULL){
    if(!isdigit(drp->d_name[0])){ continue; }

    tid = atoi(drp->d_name);

    tlist = add_list_pid(&pid_list, tid, pid);
    if(!tlist){ continue; }
    tlist->exist = TRUE;

    if(read_pid_stats(tid, tlist, &thr_nr, pid, curr)){
      // Thread doesn't exist
      tlist->exist = FALSE;
    }
  }

  closedir(dir);
}


void resource_tracker::read_stats(int curr)
{
  constexpr static char PROC[] = "/proc";
  DIR *dir;
  struct dirent *drp;
  unsigned int thr_nr;
  pid_t pid;
  struct st_pid *plist;
  struct stats_cpu *st_cpu;

  // Allocate two structures for CPU statistics.
  // No need to init them (done by read_stat_cpu() function).
  if((st_cpu = (struct stats_cpu *) malloc(sizeof(struct stats_cpu) * 2)) == NULL){
    perror("malloc");
    //exit(4);
    return;
  }
  // Read statistics for CPUs "all"
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->read_stat_cpu(st_cpu, 1);
    delete res;
  }

  // Compute the total number of jiffies spent by all processors.
  // NB: Don't add cpu_guest/cpu_guest_nice because cpu_user/cpu_nice already include them.
  tot_jiffies[curr] = st_cpu->cpu_user + st_cpu->cpu_nice +
          st_cpu->cpu_sys + st_cpu->cpu_idle +
          st_cpu->cpu_iowait + st_cpu->cpu_hardirq +
          st_cpu->cpu_steal + st_cpu->cpu_softirq;
  free(st_cpu);

  if(DISPLAY_ALL_PID(pidflag)){
    // Open /proc directory
    if((dir = opendir(PROC)) == NULL){
      perror("opendir");
      //exit(4);
      return;
    }

    // Get directory entries
    while((drp = readdir(dir)) != NULL){
      if(!isdigit(drp->d_name[0])){ continue; }

      pid = atoi(drp->d_name);

      plist = add_list_pid(&pid_list, pid, 0);
      if(!plist){ continue; }
      plist->exist = TRUE;

      if(read_pid_stats(pid, plist, &thr_nr, 0, curr)){
        // PID has terminated
        plist->exist = FALSE;

      }
      else if(DISPLAY_TID(pidflag)){
        // Read stats for threads in task subdirectory
        read_task_stats(pid, plist, curr);
      }
    }

    closedir(dir);
  }

  else{
    // Read stats for each PID in the list
    for(plist = pid_list; plist != NULL; plist = plist->next){
      // Ignore TIDs.
      // The user can only enter PIDs on the command line.
      // If there is a TID then this is because the user has used option -t,
      // and the TID has been inserted in the list by read_task_stats() function below.
      if(plist->tgid){ continue; }

      if(read_pid_stats(plist->pid, plist, &thr_nr, 0, curr)){
        // PID has terminated
        plist->exist = FALSE;
      }
      else{
        plist->exist = TRUE;

        if(DISPLAY_TID(pidflag)){
          read_task_stats(plist->pid, plist, curr);
        }
      }
    }
  }

  // Free unused PID structures
  sfree_pid(&pid_list, FALSE);
}


int resource_tracker::get_pid_to_display(int prev, int curr, unsigned int activity, unsigned int pflag,
           struct st_pid *plist)
{
  struct pid_stats *pstc = plist->pstats[curr], *pstp = plist->pstats[prev];

  // PID no longer exists
  if(!plist->exist){ return 0; }

  if(!plist->tgid){
    // This is group leader: Set it as not displayed by default
    plist->flags &= ~F_PID_DISPLAYED;
  }

  if((DISPLAY_ALL_PID(pidflag) || DISPLAY_TID(pidflag)) && DISPLAY_ACTIVE_PID(pidflag)){
    int isActive = FALSE;

    // Check that it's an "active" process
    if(DISPLAY_CPU(activity)){
      // User time already includes guest time
      if((pstc->utime != pstp->utime) || (pstc->stime != pstp->stime)){
        isActive = TRUE;
      }
      else{
        // Process is not active but if we are showing child stats then we need to look there.
        if(DISPLAY_CHILD_STATS(pflag)){
          // User time already includes guest time
          if((pstc->cutime != pstp->cutime) || (pstc->cstime != pstp->cstime)){
            isActive = TRUE;
          }
        }
      }
    }

    // If PID isn't active for any of the activities then return
    if(!isActive){ return -1; }
  }

  plist->flags |= F_PID_DISPLAYED;
  return 1;
}


int resource_tracker::write_pid_task_cpu_stats(int prev, int curr, int disp_avg,
           char *prev_string, char *curr_string, uint64_t itv, uint64_t deltot_jiffies)
{
  //printf("\nwrite_pid_task_cpu_stats\n");
  struct pid_stats *pstc, *pstp;
  struct st_pid *plist;
  int again = 0;

  //////if(dis){
  //////  PRINT_ID_HDR(prev_string, pidflag);
  //////  printf("    %%usr %%system  %%guest   %%wait    %%CPU   CPU  Command\n");
  //////}

  int i=0;
  for(plist = pid_list; plist != NULL; plist = plist->next){
    if(get_pid_to_display(prev, curr, P_A_CPU, P_TASK, plist) <= 0){ continue; }
    m_pid = plist->pid;

    pstc = plist->pstats[curr];
    pstp = plist->pstats[prev];

    ////
    m_user   = (pstc->utime - pstc->gtime) < (pstp->utime - pstp->gtime) ? 0.0 :
         sp_value(pstp->utime - pstp->gtime, pstc->utime - pstc->gtime, itv * hz / 100);
    m_system = sp_value(pstp->stime, pstc->stime, itv * hz / 100);
    m_guest  = sp_value(pstp->gtime, pstc->gtime, itv * hz / 100);
    m_wait   = sp_value(pstp->wtime, pstc->wtime, itv * hz / 100);
    m_total  = sp_value(pstp->utime + pstp->stime, pstc->utime + pstc->stime, itv * hz / 100);
    //printf("m_total: %f\n", m_total);
    if(dprint){ printf("CPU total usage: %.3f%%\n", m_total); }

    //pstc->processor // Current CPU No.

    again = 1;
  }

  return again;
}




//////////////////////////

int resource_tracker::write_stats_core(int prev, int curr, int disp_avg,
                     char *prev_string, char *curr_string)
{
  uint64_t itv, deltot_jiffies;
  int again = 0;

  // Total number of jiffies spent on the interval
  deltot_jiffies = get_interval(tot_jiffies[prev], tot_jiffies[curr]);

  itv = get_interval(uptime_cs[prev], uptime_cs[curr]);

  again += write_pid_task_cpu_stats(prev, curr, disp_avg, prev_string, curr_string, itv, deltot_jiffies);

  if(DISPLAY_ALL_PID(pidflag)){
    again = 1;
  }

  return again;
}


void resource_tracker::write_stats_avg(int curr)
{
  constexpr static char string[] = "Average:";
  write_stats_core(2, curr, TRUE, (char *) string, (char *) string);
}


int resource_tracker::write_stats(int curr)
{
  constexpr static size_t TIMESTAMP_LEN = 64;
  char cur_time[2][TIMESTAMP_LEN];

  // Get previous timestamp
  if(is_iso_time_fmt()){
    strftime(cur_time[!curr], sizeof(cur_time[!curr]), "%H:%M:%S", &ps_tstamp[!curr]);
  }
  else{
    strftime(cur_time[!curr], sizeof(cur_time[!curr]), "%X", &ps_tstamp[!curr]);
  }

  // Get current timestamp
  if(is_iso_time_fmt()){
    strftime(cur_time[curr], sizeof(cur_time[curr]), "%H:%M:%S", &ps_tstamp[curr]);
  }
  else{
    strftime(cur_time[curr], sizeof(cur_time[curr]), "%X", &ps_tstamp[curr]);
  }

  return (write_stats_core(!curr, curr, FALSE, cur_time[!curr], cur_time[curr]));
}

void resource_tracker::timer()
{
  if(timer_th != nullptr){
    timer_update.store(true);
    timer_th->detach();
    delete timer_th;
    timer_th = nullptr;
    // Make sure that the timer that is already running has stopped
    // before starting the new timer.
    while(timer_update.load()){ std::this_thread::sleep_for(std::chrono::milliseconds(100)); }
  }
  timer_th = new std::thread([this]{
    const std::chrono::milliseconds itv(((this->interval < 1) ? 1 : this->interval) * 1000);
    //printf("[DEBUG] timer started. interval=%ld, itv=%ld\n", this->interval, itv);
    while(this->loop_continue.load() && !this->timer_update.load()){
      auto start = std::chrono::system_clock::now();
      this->loop_pause.store(false);
      auto end = std::chrono::system_clock::now();
      auto waste = end - start;
      if(waste < itv){
        std::this_thread::sleep_for(itv - waste);
      }
    }
    this->timer_update.store(false);
    //printf("[DEBUG] timer stoped.\n");
  });
}

void resource_tracker::pause()
{
  while(loop_pause.load()){
    std::this_thread::sleep_for(std::chrono::milliseconds(100));
  }
  loop_pause.store(true);
}

void resource_tracker::runner()
{
  if(runner_th != nullptr){
    runner_th->detach();
    delete runner_th;
    runner_th = nullptr;
  }
  runner_th = new std::thread([this]{
    int curr = 1;
    int again;
    struct st_pid *plist;

    // Read system uptime
    {
      multiplatform::read_statistics *res = new multiplatform::read_statistics();
      res->read_uptime(&uptime_cs[0]);
      read_stats(0);
      delete res;
    }

    // Save the first stats collected. Will be used to compute the average
    this->ps_tstamp[2] = this->ps_tstamp[0];
    this->tot_jiffies[2] = this->tot_jiffies[0];
    this->uptime_cs[2] = this->uptime_cs[0];
    for(plist = pid_list; plist != NULL; plist = plist->next){
      memcpy(plist->pstats[2], plist->pstats[0], sizeof(struct pid_stats));
    }

    // Wait for SIGALRM (or possibly SIGINT) signal
    this->pause();

    // SIGINT/SIGCHLD signals caught during first interval: Exit immediately
    if(!this->loop_continue.load() && interval){ return; }

    do {
      // Every PID is potentially nonexistent
      this->set_pid_nonexistent(pid_list);

      // Get time
      get_localtime(&ps_tstamp[curr], 0);

      // Read system uptime (in 1/100th of a second)
      {
        multiplatform::read_statistics *res = new multiplatform::read_statistics();
        res->read_uptime(&(uptime_cs[curr]));
        delete res;
      }

      // Read stats
      this->read_stats(curr);

      // Print results
      again = this->write_stats(curr);

      if(!again){ return; }

      if(this->count > 0){ this->count--; }

      if(this->count){
        this->pause();
        if(!this->loop_continue.load()){
          this->count = 0;
          printf("\n");  // Skip "^C" displayed on screen
        }
        else{
          curr ^= 1;
        }
      }
    }
    while (this->count);

    // The one line format uses a raw time value rather than time strings
    // so the average doesn't really fit.

    // Write stats average
    this->write_stats_avg(curr);
  });
}


/*
void resource_tracker::rw_pidstat_loop()
{
  int curr = 1;
  int again;
  //uint64_t lines = rows;
  struct st_pid *plist;

  // Read system uptime
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->read_uptime(&uptime_cs[0]);
    read_stats(0);
    delete res;
  }

  // Set a handler for SIGALRM
  //memset(&alrm_act, 0, sizeof(alrm_act));
  //alrm_act.sa_handler = alarm_handler;
  //sigaction(SIGALRM, &alrm_act, NULL);
  //if(interval){
  //  alarm(interval);
  //}

  // Save the first stats collected. Will be used to compute the average
  ps_tstamp[2] = ps_tstamp[0];
  tot_jiffies[2] = tot_jiffies[0];
  uptime_cs[2] = uptime_cs[0];
  for(plist = pid_list; plist != NULL; plist = plist->next){
    memcpy(plist->pstats[2], plist->pstats[0], sizeof(struct pid_stats));
  }

  // Set a handler for SIGINT
  //////memset(&int_act, 0, sizeof(int_act));
  //////int_act.sa_handler = int_handler;
  //////sigaction(SIGINT, &int_act, NULL);

  // Wait for SIGALRM (or possibly SIGINT) signal
  pause();

  // SIGINT/SIGCHLD signals caught during first interval: Exit immediately
  //if(signal_caught && interval){ return; }
  if(!loop_continue.load() && interval){ return; }

  do {
    // Every PID is potentially nonexistent
    set_pid_nonexistent(pid_list);

    // Get time
    get_localtime(&ps_tstamp[curr], 0);

    // Read system uptime (in 1/100th of a second)
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->read_uptime(&(uptime_cs[curr]));
    delete res;

    // Read stats
    read_stats(curr);

    // Print results
    again = write_stats(curr);

    if(!again){ return; }

    if(count > 0){ count--; }

    if(count){
      pause();
      if(!this->loop_continue.load()){
        // SIGINT/SIGCHLD signals caught => Display average stats
        count = 0;
        printf("\n");  // Skip "^C" displayed on screen
      }
      else{
        curr ^= 1;
      }
    }
  }
  while (count);

  // The one line format uses a raw time value rather than time strings
  // so the average doesn't really fit.

  // Write stats average
  write_stats_avg(curr);
}
*/

//////////
// from common.cpp

time_t get_localtime(struct tm *rectime, int d_off)
{
  constexpr static int SEC_PER_DAY = 3600 * 24;
  time_t timer;

  timer = time(NULL);
  timer -= SEC_PER_DAY * d_off;
  localtime_r(&timer, rectime);

  return timer;
}






// ***************************************************************************
// * Main entry to the pidstat program.
// ***************************************************************************
/*
int resource_tracker::standalone_test_main(int argc, char **argv)
{
  int pid_nr = 0;
  struct utsname header;

  // Get HZ
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->get_HZ();
    hz = res->_hz;
    delete res;
  }

  // Process args...
  actflag |= P_A_CPU;
  pidflag |= P_D_UNIT;
  pidflag |= P_D_PID; // pid mode
  tskflag |= P_TASK; // task mode
  add_list_pid(&pid_list, getpid(), 0); // self pid
  interval = 1; // per 1 sec
  count = -1; // infinite loop

  // Interval not set => display stats since boot time
  if(interval < 0){ interval = 0; }

  // Count nb of proc
  //cpu_nr = get_cpu_nr(~0, FALSE);
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    cpu_nr = res->get_cpu_nr(~0, FALSE);
    delete res;
  }

  // Get time
  get_localtime(&(ps_tstamp[0]), 0);

  // Don't buffer data if redirected to a pipe.
  // Note: With musl-c, the behavior of this function is undefined except
  // when it is the first operation on the stream.
  setbuf(stdout, NULL);

  // Get system name, release number and hostname
  uname(&header);
  //print_gal_header(&(ps_tstamp[0]), header.sysname, header.release,
  //                   header.nodename, header.machine, cpu_nr, PLAIN_OUTPUT);
  std::string tmp_date;
  {
    if(&(ps_tstamp[0]) == NULL){
      tmp_date = "?/?/?";
    }
    else if(is_iso_time_fmt()){
      char tmp[64];
      strftime(tmp, 64, "%Y-%m-%d", &(ps_tstamp[0]));
      tmp_date = std::string(tmp);
    }
    else{
      char tmp[64];
      strftime(tmp, 64, "%x", &(ps_tstamp[0]));
      tmp_date = std::string(tmp);
    }
  }

  printf("%s %s (%s) \t%s \t_%s_\t(%d CPU)\n",
         header.sysname, header.release, header.nodename,
         tmp_date.c_str(), header.machine, cpu_nr);


  // Main loop
  rw_pidstat_loop();

  // Free structures
  sfree_pid(&pid_list, TRUE);

  return 0;
}
*/


int resource_tracker::stats_init(pid_t pid)
{
  int pid_nr = 0;
  struct utsname header;

  // Get HZ
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->get_HZ();
    hz = res->_hz;
    delete res;
  }

  // Process args...
  actflag |= P_A_CPU;
  pidflag |= P_D_UNIT;
  pidflag |= P_D_PID; // pid mode
  tskflag |= P_TASK; // task mode
  add_list_pid(&pid_list, pid /*getpid()*/, 0); // self pid
  //interval = 1; // per 1 sec
  count = -1; // infinite loop

  // Interval not set => display stats since boot time
  if(interval < 0){ interval = 0; }

  // Count nb of proc
  //cpu_nr = get_cpu_nr(~0, FALSE);
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    cpu_nr = res->get_cpu_nr(~0, FALSE);
    delete res;
  }

  // Get time
  get_localtime(&(ps_tstamp[0]), 0);

  // Don't buffer data if redirected to a pipe.
  // Note: With musl-c, the behavior of this function is undefined except
  // when it is the first operation on the stream.
  setbuf(stdout, NULL);

  // Get system name, release number and hostname
  uname(&header);
  std::string tmp_date;
  {
    if(&(ps_tstamp[0]) == NULL){
      tmp_date = "?/?/?";
    }
    else if(is_iso_time_fmt()){
      char tmp[64];
      strftime(tmp, 64, "%Y-%m-%d", &(ps_tstamp[0]));
      tmp_date = std::string(tmp);
    }
    else{
      char tmp[64];
      strftime(tmp, 64, "%x", &(ps_tstamp[0]));
      tmp_date = std::string(tmp);
    }
  }

  printf("%s %s (%s) \t%s \t_%s_\t(%d CPU)\n",
         header.sysname, header.release, header.nodename,
         tmp_date.c_str(), header.machine, cpu_nr);

  return 0;
}

/*
int resource_tracker::stats_run()
{
  // Main loop
  rw_pidstat_loop();
}
*/

void resource_tracker::update_interval(int64_t itv){ interval = itv; }
void resource_tracker::update_logging(bool enable){ dprint = enable; }

double resource_tracker::get_cpu_usage()
{
  if(dprint){
    printf("[DEBUG] resource_tracker::get_cpu_usage()\n[DEBUG] CPU total usage: %.3f\n", m_total);
  }
  return m_total;
}

void resource_tracker::stats_stop_request()
{
  count = 0;
  loop_continue = false;
  if(timer_th != nullptr){ timer_th->detach(); delete timer_th; timer_th = nullptr; }
  if(runner_th != nullptr){ runner_th->detach(); delete runner_th; runner_th = nullptr; }
}

int resource_tracker::stats_free()
{
  // Free structures
  sfree_pid(&pid_list, TRUE);

  return 0;
}



} // namespace multiplatform








resource_monitor::resource_monitor(pid_t pid)
{
  //enable_feature = true;
  m_pid = pid;
  rt = nullptr;

  //rt = new multiplatform::resource_tracker();
  //rt->timer();
  //rt->stats_init(pid);
  //rt->runner();
  //rt->stats_run();

//  multiplatform::stats_init(pid);
//  multiplatform::stats_run();
}

resource_monitor::~resource_monitor()
{
  if(rt != nullptr){
    rt->stats_stop_request();
    rt->stats_free();
    delete rt;
    rt = nullptr;
  }
//  multiplatform::stats_stop_request();
//  multiplatform::stats_free();
}

void resource_monitor::set_mode(bool l_enable_cut_off, bool l_enable_resource_logging,
                        int l_check_interval,
                        int l_cpu_usage_threshold, int l_memory_usage_threshold)
{
  if(l_enable_cut_off && rt == nullptr){
    rt = new multiplatform::resource_tracker();
    rt->timer();
    rt->stats_init(m_pid);
    rt->runner();
  }
  else if(!l_enable_cut_off && rt != nullptr){
    rt->stats_stop_request();
    rt->stats_free();
    delete rt;
    rt = nullptr;
  }
  //multiplatform::update_logging(l_enable_resource_logging);
  //multiplatform::update_interval((l_check_interval < 0) ? 0 : l_check_interval);
  if(rt != nullptr){
    rt->update_logging(l_enable_resource_logging);
    rt->update_interval((l_check_interval < 1) ? 1 : l_check_interval);
    rt->timer(); // Restart the timer after updating the interval parameters.
  }
  cpu_thresh = l_cpu_usage_threshold;
  mem_thresh = l_memory_usage_threshold;
}

bool resource_monitor::is_cpu_warning()
{
  //double usage = multiplatform::get_cpu_usage();
  if(rt != nullptr){
    double usage = rt->get_cpu_usage();
    return (usage >= (double) cpu_thresh ? true : false);
  }
  return false;
}

//void resource_monitor::run()
//{
//  multiplatform::stats_run();
//}



#ifndef USE_STATS

// ***************************************************************************
// * Main entry to the pidstat program.
// ***************************************************************************
int main(int argc, char **argv)
{
  //return multiplatform::standalone_test_main(argc, argv);
  //multiplatform::stats_init(getpid());
  //multiplatform::stats_run();
  //return multiplatform::stats_free();
  
  {
    resource_monitor *rm = new resource_monitor(getpid());
    std::this_thread::sleep_for(std::chrono::milliseconds(10000));
    if(rm){ delete rm; }
  }
  return 0;
}

#endif // USE_STATS


